--[[
 Automatic Belt Direction
 A Factorio mod
 Author: Magic Power
--]]

--[[    
This mod will automatically rotate belts placed by the player, according to some rules. A player's preference to belts is respected above all else, if possible.

Rule 1: Belts follow whatever direction you lay belts in.
Rule 2: Belts in between other belts will automatically turn right direction.
--]]

--[[
Technical implementation:

For rule 1, the previous location a belt was layed is remembered. If the previous and the current belt are layed adjacent, then those two belts will be turned to the direction they were layed in. 

For rule 2, a belt will be considered auto-layable if around that belt, there are belts placed in such a way that there is only one way for the middle belt to be positioned. It will then be rotated accordingly.

There are 3 exceptions to rule 1 and 2, where the belts should not be rotated. These are 3 actions where the player acts upon belts in a way he already has intentions for those belts.
   
Exception 1: Hand-rotating. Where you hold an entity in hand, and rotate it while it is still in your hand with the default 'r' key. While it is not possible to catch an event of a rotation in hand, instead we look at the previous build item and the current build item. If there was a rotation in hand in between, the difference between these items tells a hand-rotation has happened.

Exception 2: Entity-rotating. Where you hover over an ingame-entity and press the default 'r' key to rotate it. There is an event to catch the entity-rotating.

Exception 3: Overlay-rotating. Where you place an entity over an already placed entity, but rotated. Ingame this is actually done by removing the old entity, and then placing the new one. To catch these we remember the location of a removed entity by the player. If the very next placed entity is at the exact same location, we assume an overlay-rotation has occured. 


If we were to rotate a belt immediately after placing it, there would be a problem. This is because auto-rotating a belt and hand-placing a belt together can collide with each other, making coding it and using it a bit harder. Instead, the rotation of the entities is done delayed, until previous entity is no longer selected by any player.
--]]



require("funcs")

-- mydebug      = require("mydebug")

shadow       = require("shadow")
playerstatus = require("playerstatus")
memory       = require("memory")
timer        = require("timer")


-- Set timer interval at 10 ticks ~ 200 ms.
timer.setinterval(10)




-- Function BeltHasExceptions(player)
--[[
Catch everything that prevents any belt rotating. Returns true if there is anything preventing automatic belt rotating, false otherwise. 'player' contains the playerstatus table of the player that fired the build event.
--]]
function BeltHasExceptions(player)

  -- Only handle belts placed.
  if player.curr_belt.type ~= 'transport-belt' then return true end

  -- If the belt is hand-rotated (exception 1), then do not change the belt.
  if player.prev_original_direction ~= player.curr_original_direction then
    return true
  end

  -- If the belt is overlay-rotated (exception 3), then do not change the belt.
  -- Only a very short time (< 3 ticks) is overlay-rotation, any more is human interaction.
  if IsSameLocation( player.curr_belt.position, player.curr_belt.surface.index,
       player.prev_belt_removed.position, player.prev_belt_removed.surface.index ) and
     game.tick - player.prev_belt_removed.gametick < 3 
  then 
    return true 
  end

  -- A belt that has been entity-rotated (exception 2), has already been handled 
  -- by the 'on_player_rotated_entity' event.

  -- No exceptions are found
  return false
  
end -- BeltHasExceptions(player)




-- Function BeltHasRule2(player)
--[[
Check if Rule 2 applies: Belts in between other belts will automatically turn right direction. If so, rotate the belt. For rule 2, the following situations trigger a belt rotation:
   
- A belt is surrounded by 4 adjacent belts, at least one of those belts points inward, and exactly one adjacent belt points outward. The belt is then rotated towards the outward pointing belt.
   
- A belt is inbetween 2 belts that form either horizontally or vertically an ongoing belt. That belt is then rotated in the same direction as the other two. For example, if the belt above and below point south, then the middle belt will be rotated to south as well.
   
'player' contains the playerstatus table of the player that fired the build event.
Return true if a delayed rotating has been added to the memory, false otherwise
--]]
function BeltHasRule2(player)

  -- Find all entities around the current belt.
  local x = player.curr_belt.position.x
  local y = player.curr_belt.position.y
  local nr_belts = 0
  local nr_in = 0
  local nr_out = 0
  local dir_in = nil
  local dir_out = nil
  local direction = {}
  local entities = game.surfaces[ player.curr_belt.surface.index ].find_entities({{ x-1, y-1 },{ x+1, y+1}})

  -- For each belt found around the current entity: Determine if it's pointing towards 
  -- the current entity or away from it, and count the occurences.
  for _, entity in pairs(entities) do

    if entity and entity.type == 'transport-belt' then

      local dir = Adjacency(player.curr_belt.position, entity.position)

      if dir ~= nil then 
        -- Adjacent belts
        nr_belts = nr_belts + 1
        direction[dir] = entity.direction
      end
        
      if entity.direction == dir then
        -- Belt pointing outwards
        nr_out = nr_out + 1
        dir_out = dir
      end

      if entity.direction == InvDirection(dir) then
        -- Belt pointing inwards
        nr_in = nr_in + 1
        dir_in = dir
      end
    end  
  end

  -- The current belt will be auto-rotated if it is surrounded by 4 belts, there is 
  -- at least one inward pointing belt and precisely one outward pointing belt.
  if nr_in >= 1 and nr_out == 1 and nr_belts >= 4 then
    -- This belt will be delayed rotated in direction 'dir_out'.
    memory.add( player.curr_belt.name, player.curr_belt.position,
                player.curr_belt.surface.index, dir_out)
    return true
  end

  
  -- The current belt will be auto-rotated also if there is precisely 
  -- one inward pointing belt and precisely one outward pointing belt, but those belts 
  -- have to point in the same direction (e.g. horizontally or vertically).
  if  ( direction[defines.direction.north] == defines.direction.north or
        direction[defines.direction.north] == defines.direction.south ) and
     direction[defines.direction.north] == direction[defines.direction.south] then
    memory.add( player.curr_belt.name, player.curr_belt.position,
                player.curr_belt.surface.index, direction[defines.direction.north] )
    return true
  end
    
  if ( direction[defines.direction.east] == defines.direction.east or
       direction[defines.direction.east] == defines.direction.west ) and
     direction[defines.direction.east] == direction[defines.direction.west] then
    memory.add( player.curr_belt.name, player.curr_belt.position,
                player.curr_belt.surface.index, direction[defines.direction.east] )
    return true
  end

  -- The entity will not be rotated.
  return false
  
end -- BeltHasRule2(player)




-- Function BeltHasRule1(player)
--[[
Check if Rule 1 applies: Belts follow whatever direction you lay belts in. For rule 1, the previous location a belt was layed is remembered. If the previous and the current belt are layed adjacent, then those two belts will be turned to the direction they were layed in.

'player' contains the playerstatus table of the player that fired the build event.
Returns true if the belt is part of automated direction laying, false otherwise.
--]]
function BeltHasRule1(player)

  -- The previous item had to be a belt.
  if shadow.isempty( player.prev_belt ) then return false end

  -- Belts on different surfaces are ignored.
  if player.prev_belt.surface.index ~= player.curr_belt.surface.index then return false end
     
  -- The belts have to be adjacent.
  -- Also, this directly determines what direction the belts are being layed.
  local dir = Adjacency( player.prev_belt.position , player.curr_belt.position )
  if dir == nil then return false end

  -- Both belts will be delayed rotated in direction 'dir'.
  memory.add( player.prev_belt.name, player.prev_belt.position, 
              player.prev_belt.surface.index, dir)
  memory.add( player.curr_belt.name, player.curr_belt.position, 
              player.curr_belt.surface.index, dir)

  -- Automatic Belt laying is activated.
  return true

end -- BeltHasRule1(player)




--[[
Event:       on_build_entity
Description: When a player builds an entity

Prepare rotating belts when Rule 1 or Rule 2 (see above) are in effect, except when the Exceptions (see above) prevent them. The actual rotating is done in the on_tick event.
--]]    
script.on_event(defines.events.on_built_entity, function(event)

  -- If playerstatus does not exist, create it.
  if not playerstatus[event.player_index] then
    playerstatus.addplayer(event.player_index)
  end

  -- Handle event per player.
  local plr = playerstatus[event.player_index]

  -- Validate data. This looks if saved data is no longer correct, and removes it.
  ValidateData( event.created_entity.position, event.created_entity.surface.index )
      
  -- Store current belt characteristics for the given player.
  shadow.copy_entity( event.created_entity , plr.curr_belt )
  plr.curr_original_direction = plr.curr_belt.direction

  local exceptions, rule1, rule2 = false, false, false

  -- Search for exceptions, where the current entity will not be rotated.
  -- In case of exceptions, skip all further automatic belt rotating.
  exceptions = BeltHasExceptions( plr )

  -- Check if Rule 2 applies: Belts in between other belts will automatically 
  -- turn right direction. Rule 2 has precedence over rule 1. Apply Rule 2 if it does.
  if not exceptions then
    rule2 = BeltHasRule2( plr )
  end

  -- Check if Rule 1 applies: Belts follow whatever direction you lay belts in.
  -- Apply Rule 1 if it does.
  if not exceptions and not rule2 then
    rule1 = BeltHasRule1( plr )
  end

  -- See if we need to remember the belt layed.
  if not exceptions and not rule2 then
    shadow.copy_entity( plr.curr_belt , plr.prev_belt )
  else
    shadow.clear_entity( plr.prev_belt )
  end

  -- Remember the original direction of entities build
  plr.prev_original_direction = plr.curr_original_direction
  
  -- Clear the removed belt check for overlayed entities (exception 3).
  ClearPosition( plr.prev_belt_removed.position )
  plr.prev_belt_removed.surface.index = 0
  plr.prev_belt_removed.gametick = 0

  -- For some reason an 'on_tick' event directly after an 'on_built_entity' event
  -- produces irregular belt effects. This delays a possibly closeby 'on_tick' event.
  if not timer.isfinished() and timer.counter < 3 then
    timer.counter = 3
  end
  
  -- All delayed rotations are applied after the build event, and when the player is not
  -- hovering over or interacting with the memory locations.
  if timer.isfinished() and not memory.isempty() then 
    timer.activate() 
  end
  
end) -- on_event(defines.events.on_built_entity, function(event)

                 

--[[
Event:       on_player_joined_game
Description: When a player joins the game

Create the relevant status variables per player.
--]]    
script.on_event(defines.events.on_player_joined_game, function(event)
  
  -- Player joined, create a playerstatus table for him.
  playerstatus.addplayer(event.player_index)
                
end) -- on_event(defines.events.on_player_joined_game, function(event)


                 

--[[
Event:       on_player_left_game
Description: When a player leaves the game

Remove the relevant status variables per player.
--]]    
script.on_event(defines.events.on_player_left_game, function(event)
  
  -- Player left, remove his playerstatus table.
  playerstatus.removeplayer(event.player_index)
                
end) -- on_event(defines.events.on_player_left_game, function(event)


                 

--[[
Event:       on_tick
Description: Per game tick event

Execute delayed rotations. A timer specifies the intervals between checks for this, and is defined in the Timer variable. rotations are only applied if no player currently is hovering over or editing the accompanying belt.
--]]    
script.on_event(defines.events.on_tick, function(event)

  -- If there is a timer active, count it down.
  if timer.counter > 0 then 
    timer.counter = timer.counter - 1
    
    -- When the Timer reaches 1, rotate all belts in memory.
    if timer.counter == 1 then
      memory.rotateall()
      
      -- If there are still memories left, rerun the timer again.
      if not memory.isempty() then
        timer.activate()
      end
    end
  end  
end) -- on_event(defines.events.on_tick, function(event)




--[[
Event:       on_pre_player_mined_item
Description: The entity that a player removes from the game

For Exception 3: Overlay-rotating: For an entity overlay to be detected, we need to remember the previously removed belt.

Remove delayed rotated memories and the prev_belt at this location.
--]]    
script.on_event(defines.events.on_pre_player_mined_item, function(event)

  -- If playerstatus does not exist, create it.
  if not playerstatus[event.player_index] then
    playerstatus.addplayer(event.player_index)
  end
                
  local player = playerstatus[event.player_index]
  local entity = event.entity

  -- If the entity was a belt, safe the location of the to-be-removed belt. 
  -- This is needed for exception 3.
  if entity.type == 'transport-belt' then
    CopyPosition( entity.position, player.prev_belt_removed.position )
    player.prev_belt_removed.surface.index = entity.surface.index
    player.prev_belt_removed.gametick = game.tick
  end

  -- Validate data.
  ValidateData( entity.position, entity.surface.index )
                
end) -- on_event(defines.events.on_pre_player_mined_item, function(event)




--[[
Event:       on_player_rotated_entity
Description: Rotating an entity on the field

For Exception 2: Entity-rotating: Rotated entities by the player will not be changed.

Update delayed rotate memories and the PrevBelt
--]]    
script.on_event(defines.events.on_player_rotated_entity, function(event)

  local entity = event.entity

  -- Validate data.
  ValidateData( entity.position, entity.surface.index )
                
  -- Exception 2: Entity rotating has conveniently already been handled by the above action.

end) -- on_event(defines.events.on_player_rotated_entity, function(event)

                 
                 
                 
--[[
Event:       on_robot_built_entity
Description: The entity that a robot builds in the game

Update delayed rotate memories and the PrevBelt
--]]    
script.on_event(defines.events.on_robot_built_entity, function(event)

  local entity = event.created_entity

  -- Validate data.
  ValidateData( entity.position, entity.surface.index )
                
end) -- on_event(defines.events.on_robot_built_entity, function(event)

                 
                 
                 
--[[
Event:       on_robot_pre_mined
Description: The entity that a robot removes from the game

Update delayed rotate memories and the PrevBelt
--]]    
script.on_event(defines.events.on_robot_pre_mined, function(event)

  local entity = event.entity

  -- Validate data.
  ValidateData( entity.position, entity.surface.index )
                
end) -- on_event(defines.events.on_robot_pre_mined, function(event)

                 