--[[
Include file for generic functions
--]]




-- Function IsSameLocation(pos1,surface1,pos2,surface2)
--[[
Checks whether two positions are at the same location, and returns true if it is.

Takes two {x,y} position tables through pos1 and pos2, and the surfaces they are on. Returns true if pos1 and pos2 are the same location, and false otherwise.
--]]
function IsSameLocation(pos1,surface1,pos2,surface2)

  if not pos1 or not pos2 then return false end

  if surface1 ~= surface2 then return false end
  
  if ( pos1.x == pos2.x ) and ( pos1.y == pos2.y ) then
    return true
  else
    return false
  end
end -- IsSameLocation(pos1,surface1,pos2,surface2)




-- Function Adjacency(pos1,pos2)
--[[
Checks whether two positions are adjacent, and return which side the latter is.

Takes two {x,y} Position tables through pos1 and pos2. Returns how pos2 is adjacent to pos1, horizontally or vertically, or nil if they are not adjacent.
--]]
function Adjacency(pos1,pos2)
  local differenceinX = pos2.x - pos1.x
  local differenceinY = pos2.y - pos1.y

  if differenceinX == 0 then
    if differenceinY == -1 then return defines.direction.north end
    if differenceinY == 1  then return defines.direction.south end
  end
  if differenceinY == 0 then
    if differenceinX == -1 then return defines.direction.west end
    if differenceinX == 1  then return defines.direction.east end
  end

  return nil
  
end -- Adjacency(pos1,pos2)




-- Function InvDirection(dir)
--[[
Returns the inverse direction of the given direction 'dir', or nil if invalid.
--]]
function InvDirection(dir)
  if dir == defines.direction.north     then return defines.direction.south     end
  if dir == defines.direction.northeast then return defines.direction.southwest end
  if dir == defines.direction.east      then return defines.direction.west      end
  if dir == defines.direction.southeast then return defines.direction.northwest end
  if dir == defines.direction.south     then return defines.direction.north     end
  if dir == defines.direction.southwest then return defines.direction.northeast end
  if dir == defines.direction.west      then return defines.direction.east      end
  if dir == defines.direction.northwest then return defines.direction.southeast end

  return nil
end -- InvDirection(dir)




-- Function IsPlayerLocation(pos,surface)
--[[
Returns true if any player is currently at the given position, false otherwise.
--]]
function IsPlayerLocation(pos,surface)
  -- For each player, check their current location against 'pos' and 'surface'.
  for _,player in pairs(game.players) do
    if player.selected and 
        IsSameLocation( player.selected.position , player.selected.surface.index,
          pos, surface )
    then
      return true
    end
  end
  
  return false
end -- IsPlayerLocation(pos,surface)




-- Function ValidateData(pos,surface)
--[[
Validates all internal variables that refer to this position in game. Used when external code may change/remove those variables.
--]]
function ValidateData(pos,surface)
  -- For each playerstatus, remove the prev_belt entity if invalidated.
  for _,i in pairs(game.players) do

    -- If playerstatus does not exist, create it.
    if not playerstatus[i.index] then
      playerstatus.addplayer(i.index)
    end

    local player = playerstatus[i.index]

    if player and 
       IsSameLocation( player.prev_belt.position , player.prev_belt.surface.index, 
                       pos, surface ) 
    then
      shadow.clear_entity( player.prev_belt )
    end
  end

  -- Remove any memory now invalidated.
  memory.remove(pos) 

end -- ValidateData(pos,surface)




-- Function CopyPosition(from,to)
--[[
Copy a position, 'from' is the origin, 'to' is the destination.
--]]
function CopyPosition(from,to)
  to.x = from.x
  to.y = from.y
end -- CopyPosition(from,to)




-- Function ClearPosition(pos)
--[[
Clears a position (set to nil).
--]]
function ClearPosition(pos)
  pos.x = nil
  pos.y = nil
end -- ClearPosition(pos)

