--[[
Include file for the memory class functions
--]]

--[[
The memory module stores all belt rotating information, and allows for applying those belt rotations in the future.
--]]


local memory = {}




-- Function memory.setmax()
--[[
Find the last used memory and set that to be the new last memory. This function should be called every time a memory is removed.
--]]
function memory.setmax()

  -- If memory is empty, set it to 0.
  if not memory.max or memory.max == 0 then
    memory.max = 0
    return
  end
  
  -- Find the last correct memory.
  for i = memory.max, 1, -1 do
    if memory[i] then
      memory.max = i
      return
    end
  end
  
  -- Nothing found, the memory is empty.
  memory.max = 0

end -- memory.setmax()




-- Function memory.add(name,pos,surface,direction)
--[[
Add a belt to be delayed rotated, into memory. Needed are the name of the belt, the position on the map, and the direction it should be rotated to.
--]]
function memory.add(name,pos,surface,direction)

  local empty = 0

  -- Remove any memory at this location.
  memory.remove( pos, surface )
      
  -- Find the first available empty space.
  if not memory.max or memory.max == 0 then 

    -- No memory, create a new one.
    empty = 1

  else

    -- Find the first empty memory.
    for i = 1, memory.max do
      if not memory[i] then
        empty = i
        break
      end
    end
  
  end
      
  -- If still no empty save found, add it to the end.
  if empty == 0 then 
    empty = memory.max + 1
  end

  -- Write the new memory.
  memory[empty] = {}
  memory[empty].name       = name
  memory[empty].direction  = direction
  memory[empty].position   = {}
  memory[empty].position.x = pos.x
  memory[empty].position.y = pos.y
  memory[empty].surface = {}
  memory[empty].surface.index = surface
  
  -- Update the max nr of memories, if applicable.
  if not memory.max or empty > memory.max then
    memory.max = empty
  end

end -- memory.add(name,pos,surface,direction)




-- Function memory.remove(pos,surface)
--[[
Remove a belt from memory. The belt removed from the memory is any memory that has the same location as 'pos'.
--]]
function memory.remove(pos,surface)

  -- If there is no memory, then we're done.
  if not memory.max or memory.max == 0 then return end

  -- Find any memory whose location is the same as 'pos' on 'surface', and invalide it.
  for i = 1, memory.max do

    if memory[i] and IsSameLocation(memory[i].position, memory[i].surface.index , 
       pos, surface) 
    then
      memory[i] = nil
    end
  end    

  -- Set the new max size of the memory.
  memory.setmax()

end -- memory.remove(pos,surface)




-- Function memory.isempty()
--[[
Returns true if there are no memories, false otherwise.
--]]
function memory.isempty()

  if not memory.max or memory.max == 0 then
    return true
  else
    return false
  end

end -- memory.isempty()




-- Function memory.rotatememory(mem)
--[[
This function is called internally, use memory.rotateall() instead.
   
Executes a delayed rotation from memory, if possible. Returns true if the memory should be removed, false if it stays.
--]]   
function memory.rotatememory(mem)
  
  -- Check if the belt still exists and is valid.
  local entity = game.surfaces[ mem.surface.index ].find_entity( mem.name , mem.position )
  
  if not entity or entity.type ~= 'transport-belt' then 
    return true 
  end
  
  -- If a player is currently at this location, skip the memory.
  if IsPlayerLocation( mem.position, mem.surface.index ) then 
    return false
  end
  
  -- Rotate the belt in the desired position.
  entity.direction = mem.direction
  
  -- Remember to remove memory.
  return true
  
end -- memory.rotatememory(mem)
  
  
  
  
-- Function memory.rotateall()
--[[
Executes all delayed rotations from memory, if possible.
--]]   
function memory.rotateall()

  -- Skip an empty memory.
  if not memory.max or memory.max == 0 then return end

  -- For each delayed rotation, check if it can be rotated, and clear that memory.
  for i = 1, memory.max do
    if memory[i] then 
      -- Rotate the memory, and remove it if needed.
      if memory.rotatememory( memory[i] ) == true then
        memory[i] = nil
      end
    end
  end

  -- Find the new max memory.
  memory.setmax()
      
end -- memory.rotateall()
    
    
    

-- End module memory
return memory
